/*
Created on August 10, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

BitStream allows for simple bitwise reading from an unsigned character array.
It is used to load boards from compact raw data.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef BIT_STREAM_H_INCLUDED
#define BIT_STREAM_H_INCLUDED

//A structure which allows for bitwise reading of data presented in a constant
//unsigned character array.  Its fields should not be accessed outside this
//file.
typedef struct {
	char bit;
	const unsigned char *data;
} BitStream;

//Frees the contents of the specified BitStream (but not the pointer to the
//BitStream itself)
#define freeBitStream(x)

//Returns a BitStream for the specified unsigned character array
BitStream *newBitStream(const unsigned char *data) {
	BitStream *s = (BitStream *)malloc_throw(sizeof(BitStream));
	s->data = data;
	s->bit = 7;
	return s;
}

//Reads the next bit from the specified BitStream.  Returns 1 if it is true and
//0 if it is false.
char readBit(BitStream *s) {
	char bit;
	bit = (*(s->data) >> s->bit) & 1;
	if (!(s->bit)--) {
		s->bit = 7;
		(s->data)++;
	}
	return bit;
}

//Skips a bit in the specified BitStream
//This can be optimized if there is a need to do so
/*
void skipBit(BitStream *s) {
	readBit(s);
}
*/

//Skips the specified number of bits in the specified BitStream
//This can be optimized if there is a need to do so
/*
void skipBits(BitStream *s, char bits) {
	for(bits--; bits >= 0; bits--)
		skipBit(s);
}
*/

//Reads the specified number of bits from the specified BitStream as a
//nonnegative char.  Assumes that 0 <= bits <= 7.
//This can be optimized if there is a need to do so
char readChar(BitStream *s, char bits) {
	char i;
	char output = 0;
	for(i = bits - 1; i >= 0; i--)
		output |= (readBit(s) << i);
	return output;
}

//Reads the specified number of bits from the specified BitStream as an unsigned
//char.  Assumes that 0 <= bits <= 8.
//This can be optimized if there is a need to do so
/*
unsigned char readUnsignedChar(BitStream *s, char bits) {
	unsigned char output = 0;
	char i;
	for(i = bits - 1; i >= 0; i--) {
		if (readBit(s))
			output += (1 << i);
	}
	return output;
}
*/

//Reads the specified number of bits from the specified BitStream as a
//nonnegative short.  Assumes that 0 <= bits <= 15.
//This can be optimized if there is a need to do so
/*
short readShort(BitStream *s, char bits) {
	short output = 0;
	char i;
	for(i = bits - 1; i >= 0; i--) {
		if (readBit(s))
			output += (1 << (short)i);
	}
	return output;
}
*/

//Reads the specified number of bits from the specified BitStream as an unsigned
//short.  Assumes that 0 <= bits <= 16.
//This can be optimized if there is a need to do so
/*
unsigned short readUnsignedShort(BitStream *s, char bits) {
	unsigned short output = 0;
	char i;
	for(i = bits - 1; i >= 0; i--) {
		if (readBit(s))
			output += (1 << (short)i);
	}
	return output;
}
*/












#endif
