/*
Created on August 9, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains various methods used for drawing.  Several of them are
specific to the CalcChu game board and its contents.

Everything in the game is first drawn to a separate buffer, then to the screen.
This is so that the user doesn't see the game as it's being drawn, and so that
it is easy to center the screen on the TI-92 and Voyage 200.

Critters are "tiled"; that is, a critter that "leaks" off of one side of the
screen appears on the opposite side as well.  This is accomplished by drawing
such critters multiple times, once at each location where they appear, and after
drawing all of the critters, erasing the area where portions of critters that
exceed the boundaries of the game may appear.  The buffer to which the game is
drawn has a few rows of pixels above the top, so that critters that are above
the top of the screen can be drawn easily.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef DRAW_H_INCLUDED
#define DRAW_H_INCLUDED

#include "extgraph.h"
#include "game.h"
#include "key.h"
#include "time.h"

//The draw space left above the draw buffer when drawing the board.  This
//variable could probably be made a little smaller, but I'm choosing not to
//sweat those one or two extra pixels.
#define BORDER 17
//The height of the area below the board
//(100 - (14 * HEIGHT + 1))
#define BOTTOM_BAR_HEIGHT 15
//The number of bytes of the bottom bar, the area below the board
//BOTTOM_BAR_HEIGHT * (240 / 8)
#define BOTTOM_BAR_BUFFER_SIZE 450
//The number of bytes above the bottom bar, the area below the board
//(100 - BOTTOM_BAR_HEIGHT) * (240 / 8)
#define BOTTOM_BAR_BUFFER_OFFSET 2550
//The size of the buffer in which the board is drawn
//(14 * HEIGHT + 1) * (240 / 8);
#define BOARD_DRAW_SIZE 2550
//The number of garbagy bytes in a draw buffer before the actual data (see
//drawCritter).  Corresponds to BORDER rows of the LCD screen.
//BORDER * (240 / 8)
#define EXPANDED_DRAW_OFFSET 510
//The size of a draw buffer with garbagy bytes before the actual data
#define EXPANDED_DRAW_SIZE 4350
//The upper limit of the frac parameter in drawCritter
#define MAX_FRAC 19

//The draw width of the board in pixels
//14 * WIDTH + 1;
#define DRAW_WIDTH 127
//The draw height of the board in pixels
//14 * HEIGHT + 1;
#define DRAW_HEIGHT 85





//===== Sprites =====
//The XO in some of the #define's means "x offset", while YO means "y offset"
#define CAT_NORTH ((unsigned short[]) \
	{0x1FC0, 0x3FE0, 0x7070, 0xEFB8, 0xFFF8, 0xFFF8, 0xCF98, 0xCF98, \
	 0xFAF8, 0xF778, 0x7DF0, 0x38E0, 0x1040})
#define CAT_EAST ((unsigned short[]) \
	{0x1FC0, 0x3FE0, 0x79F0, 0xF9B8, 0x6FD8, 0x37D8, 0x1FD8, 0x37D8, \
	 0x6FD8, 0xF9B8, 0x79F0, 0x3FE0, 0x1FC0})
//CAT_SOUTH is not #define'd, as it is used twice
static unsigned short CAT_SOUTH[] =
	{0x1040, 0x38E0, 0x7DF0, 0xF778, 0xFAF8, 0xCF98, 0xCF98, 0xFFF8,
	 0xFFF8, 0xEFB8, 0x7070, 0x3FE0, 0x1FC0};
#define CAT_WEST ((unsigned short[]) \
	{0x1FC0, 0x3FE0, 0x7CF0, 0xECF8, 0xDFB0, 0xDF60, 0xDFC0, 0xDF60, \
	 0xDFB0, 0xECF8, 0x7CF0, 0x3FE0, 0x1FC0})

#define MOUSE_NORTH ((unsigned char[]) \
	{0x20, 0x70, 0x70, 0xF8, 0xF8, 0xF8, 0xF8, 0x70, 0x20, 0x20, 0x10})
//MOUSE_EAST is not #define'd, as it is used twice
static unsigned short MOUSE_EAST[] = {0x0F00, 0x1FC0, 0x7FE0, 0x9FC0, 0x0F00};
#define MOUSE_SOUTH ((unsigned char[]) \
	{0x40, 0x20, 0x20, 0x70, 0xF8, 0xF8, 0xF8, 0xF8, 0x70, 0x70, 0x20})
#define MOUSE_WEST ((unsigned short[]){0x1E00, 0x7F20, 0xFFC0, 0x7F00, 0x1E00})

static unsigned char HATCH_NORTH[] = {0x70, 0xF8, 0xF8, 0xF8, 0xF8};
static unsigned char HATCH_EAST[] = {0xF0, 0xF8, 0xF8, 0xF8, 0xF0};
static unsigned char HATCH_SOUTH[] = {0xF8, 0xF8, 0xF8, 0xF8, 0x70};
static unsigned char HATCH_WEST[] = {0x78, 0xF8, 0xF8, 0xF8, 0x78};

#define PIT ((unsigned short[]) \
	{0xC018, 0xE038, 0x7070, 0x38E0, 0x1DC0, 0x0F80, 0x0700, 0x0F80, \
	 0x1DC0, 0x38E0, 0x7070, 0xE038, 0xC018})

#define ROCKET_HEIGHT 7
#define ROCKET_XO 4
#define ROCKET_YO 4

static unsigned char P1_ROCKET[ROCKET_HEIGHT] =
	{0x10, 0x28, 0x44, 0x82, 0x44, 0x28, 0x10};
static unsigned char P2_ROCKET[ROCKET_HEIGHT] =
	{0x10, 0x38, 0x7C, 0xFE, 0x7C, 0x38, 0x10};

#define SPECIAL_MOUSE_PIC ((unsigned char[]) \
	{0x7C, 0xFE, 0xC6, 0xC6, 0x0E, 0x1C, 0x18, 0x18, 0x00, 0x18, 0x18})
#define SUPER_MOUSE_PIC ((unsigned short[]) \
	{0xF9C0, 0xFBE0, 0xC360, 0xF360, 0xFB60, 0x1B60, 0x1B60, 0xFBE0, 0xF1C0})

#define ARROW_HEIGHT 11
#define ARROW_XO 2
#define ARROW_YO 2

static unsigned short P1_ARROW_NORTH[ARROW_HEIGHT] =
	{0x0400, 0x0A00, 0x1100, 0x2080, 0x4040, 0xF1E0, 0x1100, 0x1100,
	 0x1100, 0x1100, 0x1F00};
static unsigned short P1_ARROW_EAST[ARROW_HEIGHT] =
	{0x0400, 0x0600, 0x0500, 0xFC80, 0x8040, 0x8020, 0x8040, 0xFC80,
	 0x0500, 0x0600, 0x0400};
static unsigned short P1_ARROW_SOUTH[ARROW_HEIGHT] =
	{0x1F00, 0x1100, 0x1100, 0x1100, 0x1100, 0xF1E0, 0x4040, 0x2080,
	 0x1100, 0x0A00, 0x0400};
static unsigned short P1_ARROW_WEST[ARROW_HEIGHT] =
	{0x0400, 0x0C00, 0x1400, 0x27E0, 0x4020, 0x8020, 0x4020, 0x27E0,
	 0x1400, 0x0C00, 0x0400};

static unsigned short P2_ARROW_NORTH[ARROW_HEIGHT] =
	{0x0400, 0x0E00, 0x1F00, 0x3F80, 0x7FC0, 0xFFE0, 0x1F00, 0x1F00,
	 0x1F00, 0x1F00, 0x1F00};
static unsigned short P2_ARROW_EAST[ARROW_HEIGHT] =
	{0x0400, 0x0600, 0x0700, 0xFF80, 0xFFC0, 0xFFE0, 0xFFC0, 0xFF80,
	 0x0700, 0x0600, 0x0400};
static unsigned short P2_ARROW_SOUTH[ARROW_HEIGHT] =
	{0x1F00, 0x1F00, 0x1F00, 0x1F00, 0x1F00, 0xFFE0, 0x7FC0, 0x3F80,
	 0x1F00, 0x0E00, 0x0400};
static unsigned short P2_ARROW_WEST[ARROW_HEIGHT] =
	{0x0400, 0x0C00, 0x1C00, 0x3FE0, 0x7FE0, 0xFFE0, 0x7FE0, 0x3FE0,
	 0x1C00, 0x0C00, 0x0400};

#define SMALL_ARROW_HEIGHT 7
#define SMALL_ARROW_XO 4
#define SMALL_ARROW_YO 4

static unsigned char P1_SMALL_ARROW_NORTH[SMALL_ARROW_HEIGHT] =
	{0x10, 0x28, 0x44, 0xEE, 0x28, 0x28, 0x38};
static unsigned char P1_SMALL_ARROW_EAST[SMALL_ARROW_HEIGHT] =
	{0x10, 0x18, 0xF4, 0x82, 0xF4, 0x18, 0x10};
static unsigned char P1_SMALL_ARROW_SOUTH[SMALL_ARROW_HEIGHT] =
	{0x38, 0x28, 0x28, 0xEE, 0x44, 0x28, 0x10};
static unsigned char P1_SMALL_ARROW_WEST[SMALL_ARROW_HEIGHT] =
	{0x10, 0x30, 0x5E, 0x82, 0x5E, 0x30, 0x10};

static unsigned char P2_SMALL_ARROW_NORTH[SMALL_ARROW_HEIGHT] =
	{0x10, 0x38, 0x7C, 0xFE, 0x38, 0x38, 0x38};
static unsigned char P2_SMALL_ARROW_EAST[SMALL_ARROW_HEIGHT] =
	{0x10, 0x18, 0xFC, 0xFE, 0xFC, 0x18, 0x10};
static unsigned char P2_SMALL_ARROW_SOUTH[SMALL_ARROW_HEIGHT] =
	{0x38, 0x38, 0x38, 0xFE, 0x7C, 0x38, 0x10};
static unsigned char P2_SMALL_ARROW_WEST[SMALL_ARROW_HEIGHT] =
	{0x10, 0x30, 0x7E, 0xFE, 0x7E, 0x30, 0x10};





//Returns the x coordinate of the upper-left hand corner where the specified
//space is drawn.
short drawX(char pos) {
	return 14 * (pos % WIDTH);
}

//Returns the y coordinate of the upper-left hand corner where the specified
//space is drawn
short drawY(char pos) {
	return 14 * (pos / WIDTH);
}

//Sets the specified buffer to be the buffer to which the draw functions draw
void setDrawBuffer(void *buffer) {
	PortSet(buffer, 239, 127);
}

//Clears the specified draw buffer
void clearDrawBuffer(void *buffer) {
	memset(buffer, 0, LCD_SIZE);
}

//Calls clearDrawBuffer and setDrawBuffer on the specified buffer
void clearAndSetDrawBuffer(void *buffer) {
	clearDrawBuffer(buffer);
	setDrawBuffer(buffer);
}

//Calls FontSetSys(F_4x6).  Included to reduce program size.
void setFont4x6() {
	FontSetSys(F_4x6);
}

//Calls FontSetSys(F_6x8).  Included to reduce program size.
void setFont6x8() {
	FontSetSys(F_6x8);
}

//Calls FontSetSys(F_8x10).  Included to reduce program size.
void setFont8x10() {
	FontSetSys(F_8x10);
}

//Calls DrawStr(x, y, str, A_NORMAL).  Included to reduce program size.
void drawStrNormal(short x, short y, const char *str) {
	DrawStr(x, y, str, A_NORMAL);
}

//Outlines the specified rectangle
/*
This is manually inlined the one place it is used.
void outline(short x1, short y1, short x2, short y2, short attr) {
	SetCurAttr(attr);
	MoveTo(x1, y1);
	LineTo(x2, y1);
	LineTo(x2, y2);
	LineTo(x1, y2);
	LineTo(x1, y1);
}
*/

//Returns the height of the specified font
short fontHeight(short font) {
	switch (font) {
		case F_4x6:
			return 4;
		case F_6x8:
			return 6;
		//case F_8x10:
		default:
			return 8;
	}
}

#define PIT_XO 1
#define PIT_YO 1
#define PIT_HEIGHT 13
#define HATCH_XO 5
#define HATCH_YO 5
#define HATCH_HEIGHT 5

#define HORIZONTAL_GRIDLINE_HEIGHT 1
#define VERTICAL_GRIDLINE_HEIGHT 15

//Draws the specified board to the specified buffer
static void drawBoard(void *buffer, Board *board) {
	static unsigned short HORIZONTAL_GRIDLINE[] = {0x924A};
	static unsigned char VERTICAL_GRIDLINE[] =
		{0x80, 0x00, 0x00, 0x80, 0x00, 0x00, 0x80, 0x00,
		 0x00, 0x80, 0x00, 0x00, 0x80, 0x00, 0x80};
	
	static unsigned char *HATCHES[4] =
		{HATCH_NORTH, HATCH_EAST, HATCH_SOUTH, HATCH_WEST};
	
	short i, j;
	short x, y;
	short x2, y2;
	
	Hatch *hatch;
	short pos;
	char wall;
	
	setDrawBuffer(buffer);
	
	//Draw pits
	for(i = 0; i < WIDTH_TIMES_HEIGHT; i++) {
		if (board->pits[i])
			Sprite16_OR(drawX(i) + PIT_XO, drawY(i) + PIT_YO,
						PIT_HEIGHT,
						PIT,
						buffer);
	}
	
	
	
	//Draw hatches
	for(i = 0, hatch = board->hatches; i < board->numHatches; i++, hatch++) {
		pos = hatch->pos;
		Sprite8_OR(drawX(pos) + HATCH_XO, drawY(pos) + HATCH_YO,
				   HATCH_HEIGHT,
				   HATCHES[hatch->dir],
				   buffer);
	}
	
	
	
	//Draw walls
	pos = 0;
	for(j = 0; j < HEIGHT; j++) {
		y = drawY(pos);
		y2 = y + 14;
		for(i = 0; i < WIDTH; i++, pos++) {
			x = drawX(pos);
			x2 = x + 14;
			
			wall = board->walls[pos];
			
			//North
			if (wall & 1)
				DrawLine(x, y, x2, y, A_NORMAL);
			else
				Sprite16_OR(x, y,
							HORIZONTAL_GRIDLINE_HEIGHT,
							HORIZONTAL_GRIDLINE,
							buffer);
			
			//West
			if (wall >> 3)
				DrawLine(x, y, x, y2, A_NORMAL);
			else
				Sprite8_OR(x, y,
						   VERTICAL_GRIDLINE_HEIGHT,
						   VERTICAL_GRIDLINE,
						   buffer);
			
			//South
			if (j == HEIGHT - 1) {
				if ((wall >> 2) & 1)
					DrawLine(x, y2, x2, y2, A_NORMAL);
				else
					Sprite16_OR(x, y2,
								HORIZONTAL_GRIDLINE_HEIGHT,
								HORIZONTAL_GRIDLINE,
								buffer);
			}
			
			//East
			if (i == WIDTH - 1) {
				if ((wall >> 1) & 1)
					DrawLine(x2, y, x2, y2, A_NORMAL);
				else
					Sprite8_OR(x2, y,
							   VERTICAL_GRIDLINE_HEIGHT,
							   VERTICAL_GRIDLINE,
							   buffer);
			}
		}
	}
	
	PortRestore();
}

//Draws the specified arrow to the specified buffer
static void drawArrow(void *buffer,
					  char isPlayer1,
					  short pos,
					  char dir,
					  char isSmall) {
	short x, y;
	
	x = drawX(pos);
	y = drawY(pos);
	
	if (!isSmall)
		Sprite16_OR(x + ARROW_XO, y + ARROW_YO,
					ARROW_HEIGHT,
					(isPlayer1 ? (unsigned short *[4]){P1_ARROW_NORTH,
													   P1_ARROW_EAST,
													   P1_ARROW_SOUTH,
													   P1_ARROW_WEST} :
					(unsigned short *[4]){P2_ARROW_NORTH,
										  P2_ARROW_EAST,
										  P2_ARROW_SOUTH,
										  P2_ARROW_WEST})[dir],
					buffer);
	else
		Sprite8_OR(x + SMALL_ARROW_XO, y + SMALL_ARROW_YO,
				   SMALL_ARROW_HEIGHT,
					(isPlayer1 ? (unsigned char *[4]){P1_SMALL_ARROW_NORTH,
													  P1_SMALL_ARROW_EAST,
													  P1_SMALL_ARROW_SOUTH,
													  P1_SMALL_ARROW_WEST} :
					(unsigned char *[4]){P2_SMALL_ARROW_NORTH,
										 P2_SMALL_ARROW_EAST,
										 P2_SMALL_ARROW_SOUTH,
										 P2_SMALL_ARROW_WEST})[dir],
					buffer);
}

/*
//Draws arrows for the specified game to the specified buffer
static void drawArrows(void *buffer, Game *game) {
	short i;
	
	Arrow *arrow;
	
	for(i = 0, arrow = game->p1Arrows; i < MAX_ARROWS; i++, arrow++) {
		if (arrow->exists)
			drawArrow(buffer, 1, arrow->pos, arrow->dir, arrow->isSmall);
	}
	
	for(i = 0, arrow = game->p2Arrows; i < MAX_ARROWS; i++, arrow++) {
		if (arrow->exists)
			drawArrow(buffer, 0, arrow->pos, arrow->dir, arrow->isSmall);
	}
}
*/

/*
//Tiles the specified sprite within the board's draw rectangle, so that if it
//exceeds the edge, it is drawn on the opposite edge as well.  Assumes that the
//sprite at least partially overlaps the draw rectangle.
//This function is manually inlined the one place it is used.
static void tile(short x, short y,
				 short w, short h,
				 void *sprite,
				 void *buffer,
				 void (*drawFunc)(short, short, short, void *, void *)) {
	short a, b;
	if (y + h > DRAW_HEIGHT)
		y -= DRAW_HEIGHT - 1;
	if (x + w > DRAW_WIDTH)
		x -= DRAW_WIDTH - 1;
	
	if (y < 0) {
		if (x < 0) {
			a = x + DRAW_WIDTH - 1;
			b = y + DRAW_HEIGHT - 1;
			drawFunc(x + 240, y - 1 + BORDER,
					 h,
					 sprite,
					 buffer - EXPANDED_DRAW_OFFSET);
			drawFunc(x + 240, b, h, sprite, buffer);
			drawFunc(a, y + BORDER, h, sprite, buffer - EXPANDED_DRAW_OFFSET);
			drawFunc(a, b, h, sprite, buffer);
		}
		else {
			drawFunc(x, y, h, sprite, buffer);
			drawFunc(x, y + DRAW_HEIGHT - 1, h, sprite, buffer);
		}
	}
	else if (x < 0) {
		drawFunc(x + 240, y - 1, h, sprite, buffer);
		drawFunc(x + DRAW_WIDTH - 1, y, h, sprite, buffer);
	}
	else
		drawFunc(x, y, h, sprite, buffer);
}
*/

//Same as Sprite8_OR, except 4th argument is of type void * instead of unsigned
//char *
static void sprite8ForVoidStar(short x, short y, short h, void *sprite, void *buffer) {
	Sprite8_OR(x, y, h, (unsigned char *)sprite, buffer);
}

//Same as Sprite16_OR, except 4th argument is of type void * instead of unsigned
//short *
static void sprite16ForVoidStar(short x, short y, short h, void *sprite, void *buffer) {
	Sprite16_OR(x, y, h, (unsigned short *)sprite, buffer);
}

//Called after drawing some critters, to clean up the area where extra pieces of
//sprites may have been drawn
void drawCrittersPost(void *buffer) {
	short a = DRAW_HEIGHT + BORDER - 1;
	
	//Clean up the edges of the drawing, where extra pieces of sprites may have
	//been drawn
	FastFillRect(buffer, 240 - BORDER, 0, 239, a, A_REVERSE);
	FastFillRect(buffer, DRAW_WIDTH, 0, DRAW_WIDTH + BORDER - 1, a, A_REVERSE);
	FastFillRect(buffer, 0, DRAW_HEIGHT, DRAW_WIDTH - 1, a, A_REVERSE);
	
	#if 0
	//Not surprisingly, this implementation is slower
	static const unsigned char RIGHT1 = (DRAW_WIDTH + 7) / 8;
	static const unsigned char RIGHT2 = (DRAW_WIDTH + BORDER - 1 + 7) / 8;
	static const unsigned char RIGHT2_MINUS_RIGHT1 = RIGHT2 - RIGHT1;
	static const unsigned char RIGHT3 = (240 - BORDER) / 8;
	static const unsigned char RIGHT_BYTE =
		~((char)((1 << (7 - ((DRAW_WIDTH - 1) % 8))) - 1));
	static const unsigned char WIDTH_IN_BYTES = (DRAW_WIDTH - 1 + 7) / 8;
	
	unsigned char *c;
	unsigned char *end;
	
	end = (unsigned char *)(buffer + 30 * (DRAW_HEIGHT + BORDER - 1));
	
	//If the width is sufficiently small, the relevant data will be ignored, so
	//we need not clear them
	if (LCD_WIDTH > 180) {
		for(c = buffer + RIGHT3; c < end; c += 30)
			memset(c, 0, 30 - RIGHT3);
	}
	
	for(c = buffer + RIGHT1; c < end; c += 30) {
		memset(c, 0, RIGHT2_MINUS_RIGHT1);
		*(c - 1) &= RIGHT_BYTE;
	}
	
	for(c = buffer + DRAW_HEIGHT * 30; c < end; c += 30)
		memset(c, 0, WIDTH_IN_BYTES);
	#endif
	
	//PortRestore();
}

#define CAT_O 1
#define CAT_SIZE 13

#define MOUSE_XO_EVEN 5
#define MOUSE_XO_ODD 2
#define MOUSE_HEIGHT_EVEN 11
#define MOUSE_HEIGHT_ODD 5

#define SPECIAL_MOUSE_HEIGHT 11
#define SPECIAL_MOUSE_XO 4
#define SPECIAL_MOUSE_YO 2

#define SUPER_MOUSE_HEIGHT 9
#define SUPER_MOUSE_XO 2
#define SUPER_MOUSE_YO 3

//Draws the specified critter for the specified game to the specified buffer.
//Assumes that it exists.  frac indicates the number of MAX_FRAC'ths of a game
//tick that the game is beyond the last game tick that was executed.  It is used
//to allow for smooth animation between game ticks.  Note that drawCrittersPost
//must be called after drawing a group of critters.  Assumes that the buffer has
//BORDER rows above and below the usual board.  Note that buffer[0] should
//indicate the top-left corner, and buffer[x] should be accessible for some
//negative x.
static void drawCritter(void *buffer, Critter *critter, short frac) {
	short a, b;
	short x, y;
	void *sprite;
	char isChars;
	char dir;
	char offset;
	char pos;
	char h;
	void (*drawFunc)(short, short, short, void *, void *);
	
	
	dir = critter->dir;
	
	//Compute the relevant sprite
	switch (critter->type) {
		case CAT:
			sprite =
				((void *[]){CAT_NORTH, CAT_EAST, CAT_SOUTH, CAT_WEST})[dir];
			h = CAT_SIZE;
			x = CAT_O;
			y = CAT_O;
			isChars = 0;
			break;
		case MOUSE:
			if (dir & 1) {
				sprite = (dir == EAST) ? MOUSE_EAST : MOUSE_WEST;
				h = MOUSE_HEIGHT_ODD;
				x = MOUSE_XO_ODD;
				y = MOUSE_XO_EVEN;
				isChars = 0;
			}
			else {
				sprite = (dir == NORTH) ? MOUSE_NORTH : MOUSE_SOUTH;
				h = MOUSE_HEIGHT_EVEN;
				x = MOUSE_XO_EVEN;
				y = MOUSE_XO_ODD;
				isChars = 1;
			}
			
			break;
		case SPECIAL_MOUSE:
			sprite = SPECIAL_MOUSE_PIC;
			h = SPECIAL_MOUSE_HEIGHT;
			x = SPECIAL_MOUSE_XO;
			y = SPECIAL_MOUSE_YO;
			isChars = 1;
			break;
		//case SUPER_MOUSE:
		default:
			sprite = SUPER_MOUSE_PIC;
			h = SUPER_MOUSE_HEIGHT;
			x = SUPER_MOUSE_XO;
			y = SUPER_MOUSE_YO;
			isChars = 0;
	}
	
	//Add an offset based on the portion field
	//Note that this is not *exactly* "accurate" (the ideal is 14 * (portion *
	//MAX_FRAC + frac) / (MAX_PORTION * MAX_FRAC)), but it ought to be fast
	offset = (critter->portion * 19 +
			  (critter->type == CAT ? CAT_ADVANCE : MOUSE_ADVANCE) * frac) >> 3;
	
	switch (dir) {
		case NORTH:
			y -= offset;
			break;
		case EAST:
			x += offset;
			break;
		case SOUTH:
			y += offset;
			break;
		//case WEST:
		default:
			x -= offset;
	}
	
	pos = critter->pos;
	x += drawX(pos);
	y += drawY(pos);
	
	//Tile the sprite within the board's draw rectangle, so that if it exceeds
	//the edge, it is drawn on the opposite edge as well
	drawFunc = isChars ? sprite8ForVoidStar : sprite16ForVoidStar;
	
	if (y + h > DRAW_HEIGHT)
		y -= DRAW_HEIGHT - 1;
	if (x + (isChars ? 8 : 16) > DRAW_WIDTH)
		x -= DRAW_WIDTH - 1;
	
	if (y < 0) {
		if (x < 0) {
			a = x + DRAW_WIDTH - 1;
			b = y + DRAW_HEIGHT - 1;
			drawFunc(x + 240, y - 1 + BORDER,
					 h,
					 sprite,
					 buffer - EXPANDED_DRAW_OFFSET);
			drawFunc(x + 240, b, h, sprite, buffer);
			drawFunc(a, y + BORDER, h, sprite, buffer - EXPANDED_DRAW_OFFSET);
			drawFunc(a, b, h, sprite, buffer);
		}
		else {
			drawFunc(x, y, h, sprite, buffer);
			drawFunc(x, y + DRAW_HEIGHT - 1, h, sprite, buffer);
		}
	}
	else if (x < 0) {
		drawFunc(x + 240, y - 1, h, sprite, buffer);
		drawFunc(x + DRAW_WIDTH - 1, y, h, sprite, buffer);
	}
	else
		drawFunc(x, y, h, sprite, buffer);
}

//Draws critters for the specified game to the specified buffer.  catsFrac and
//miceFrac indicate the number of MAX_FRAC'ths of a game tick that the critters
//of the corresponding type are beyond the last game tick that was executed.
//There are separate values for mice and cats so that cats' motion can be paused
//for the "Cat Nap" special event, without pausing mice's motion.  Automatically
//calls drawCrittersPost.  Assumes that the buffer has BORDER rows above and
//below the usual board.  Note that buffer[0] should indicate the top-left
//corner, and buffer[x] should be accessible for some negative x.
static void drawCritters(void *buffer,
						 Game *game,
						 short catsFrac, short miceFrac) {
	Critter *critter;
	Critter *end = game->critters + MAX_CRITTERS;
	
	for(critter = game->critters; critter < end; critter++) {
		if (critter->numCritters)
			drawCritter(buffer,
						critter,
						critter->type == CAT ? catsFrac : miceFrac);
	}
	drawCrittersPost(buffer);
}

//Draws a selection rectangle at the specified position to the specified buffer
void drawSelection(void *buffer, char selectedPos) {
	//static const unsigned short SELECTION[] =
	//	{0xFFF8, 0x8008, 0x8008, 0x8008, 0x8008, 0x8008, 0x8008, 0x8008,
	//	 0x8008, 0x8008, 0x8008, 0x8008, 0xFFF8};
	
	short x = drawX(selectedPos) + 1;
	short y = drawY(selectedPos) + 1;
	
	setDrawBuffer(buffer);
	SetCurAttr(A_NORMAL);
	MoveTo(x, y);
	LineTo(x + 12, y);
	LineTo(x + 12, y + 12);
	LineTo(x, y + 12);
	LineTo(x, y);
	PortRestore();
}

//Draws the specified string at the center of the area right of the playing
//board
void drawCenterRight(short y, const char *str) {
	drawStrNormal(DRAW_WIDTH + (160 - DRAW_WIDTH -
								DrawStrWidth(str, FontGetSys())) / 2,
				  y,
				  str);
}

//Draws the players' scores
void drawScores(void *buffer, Game *game) {
	char s[10];
	
	setDrawBuffer(buffer);
	
	setFont6x8();
	drawCenterRight(19, "White");
	drawCenterRight(48, "Black");
	setFont8x10();
	sprintf(s, "%d", game->score1);
	drawCenterRight(29, s);
	sprintf(s, "%d", game->score2);
	drawCenterRight(58, s);
	
	PortRestore();
}

//Draws a rockets at the specified position
void drawRocket(void *buffer, char pos, char isPlayer1) {
	Sprite8_OR(drawX(pos) + ROCKET_XO, drawY(pos) + ROCKET_YO,
			   ROCKET_HEIGHT,
			   isPlayer1 ? P1_ROCKET : P2_ROCKET,
			   buffer);
}

//Draws the rockets for the specified game
void drawRockets(void *buffer, Game *game) {
	SpaceData *boardData = game->boardData;
	char rocketAPos = game->board->rocketAPos;
	char rocketBPos = game->board->rocketBPos;
	drawRocket(buffer, rocketAPos, boardData[rocketAPos].type == SPACE_ROCKET1);
	drawRocket(buffer, rocketBPos, boardData[rocketBPos].type == SPACE_ROCKET1);
}

//Helper function for drawMatchScore
void drawMatchScoreOrNot(void *buffer,
						 short x, short y,
						 char score,
						 char isPlayer1) {
	if (score)
		Sprite8_OR(x, y,
				   ROCKET_HEIGHT,
				   isPlayer1 ? P1_ROCKET : P2_ROCKET,
				   buffer);
	else
		DrawLine(x, y + ROCKET_HEIGHT - 1, x + ROCKET_HEIGHT - 1, y, A_NORMAL);
}

//Draws the number of games that each player has won so far to buffer, which
//stores BOTTOM_BAR_BUFFER_SIZE bytes representing the bottom of the screen
void drawMatchScore(void *buffer, short score1, short score2, short firstTo) {
	static const short space = 3;
	static const short centerX = (14 * WIDTH + 2) / 2;
	static const short centerY = (100 - (14 * HEIGHT + 1) + 1) / 2;
	
	short i;
	short dx;
	short y = centerY - (ROCKET_HEIGHT + 1) / 2;
	
	PortSet(buffer, 239, BOTTOM_BAR_HEIGHT);
	
	memset(buffer, 0, BOTTOM_BAR_BUFFER_SIZE);
	setFont6x8();
	drawStrNormal(centerX - 9, centerY - 4, "vs.");
	
	dx = 9 + space;
	if (firstTo < 5)
		dx += ROCKET_HEIGHT + space;
	for(i = firstTo - 1; i >= 0; i--) {
		drawMatchScoreOrNot(buffer,
							centerX - dx - ROCKET_HEIGHT + 1, y,
							score1 > i,
							1);
		drawMatchScoreOrNot(buffer, centerX + dx, y, score2 > i, 0);
		dx += ROCKET_HEIGHT + space;
	}
	
	PortRestore();
}

//Draws the specified number of seconds remaining to the specified buffer
void drawTime(void *buffer, short seconds) {
	char s[10];
	sprintf(s, "%d", seconds);
	setDrawBuffer(buffer);
	setFont8x10();
	drawCenterRight(87, s);
	PortRestore();
}

//Draws a message over the board.  The message may be at most 20 characters
//long.
void drawGameMessage(void *buffer, const char *message) {
	static const short h = 7;
	static const short y = (DRAW_HEIGHT - 7 + 1) / 2;
	
	short w = DrawStrWidth(message, F_6x8);
	short x = (DRAW_WIDTH - w + 1) / 2;
	
	setDrawBuffer(buffer);
	FastFillRect(buffer,
				 max(x - 4, 0), y - 5,
				 min(x + w + 4, DRAW_WIDTH - 1), y + h + 4,
				 A_REVERSE);
	setFont6x8();
	drawStrNormal(x, y, message);
	PortRestore();
}

//Draws the specified buffer to the screen, where only the top-left 160x100
//pixels have drawn content.  It centers the contents on the TI-92 and Voyage
//200 screens.
void draw89Size(void *buffer) {
	if (TI89)
		//2990 = 99 * (240 / 8) + (160 / 8)
		memcpy(LCD_MEM, buffer, 2990);
	else {
		//425 = (128 - 100) / 2 * (240 / 8) + ((240 - 160) / 2) / 8
		memset(LCD_MEM, 0, 425);
		//2990 = 99 * (240 / 8) + (160 / 8)
		memcpy(LCD_MEM + 425, buffer, 2990);
		//3415 = 425 + 2990
		memset(LCD_MEM + 3415, 0, 425);
	}
}

//Draws the specified string in the horizontal center of the screen
void drawCenter(short y, const char *s) {
	drawStrNormal((160 - DrawStrWidth(s, FontGetSys())) / 2, y, s);
}

//Draws the specified message.  Permits newline characters.  The text is
//initially drawn with the 8x10 font.  When a '|' or a '`' is encountered, it
//switches to switch to the 6x8 or Xx6 font respectively, and adds some space.
//Assumes that s is not the empty string.  Automatically switches to menu mode.
void drawMessage(const char *s) {
	short i, j;
	short y = 30;
	char buffer[LCD_SIZE];
	short h = 10;
	char c;
	char line[100];
	char newline;
	char space;
	short newFont = -1;
	short newH = -1;
	
	switchToMenuMode();
	
	clearAndSetDrawBuffer(buffer);
	setFont8x10();
	
	i = 0;
	j = 0;
	do {
		c = s[i++];
		newline = 1;
		space = 0;
		switch (c) {
			case '|':
				space = 1;
				newFont = F_6x8;
				newH = 8;
				break;
			case '`':
				space = 1;
				newFont = F_4x6;
				newH = 6;
				break;
			case '\n': case '\0':
				newFont = FontGetSys();
				newH = h;
				break;
			default:
				line[j++] = c;
				newline = 0;
				break;
		}
		
		if (newline) {
			line[j] = '\0';
			drawCenter(y, line);
			y += h + 2;
			if (space)
				y += (h / 2) + 1;
			j = 0;
			h = newH;
			FontSetSys(newFont);
		}
	}
	while (c != '\0');
	
	PortRestore();
	draw89Size(buffer);
}

//Draws the specified message.  Draws it using the 6x8 font, and centers it
//horizontally and vertically in the screen.  Automatically performs word wrap,
//adding line breaks after spaces or hyphens.  Assumes that no line is longer
//than 26 characters, and assumes that s is not the empty string.  Automatically
//switches to menu mode.
void drawMessage2(const char *s) {
	short i, j, k;
	short y;
	short yo;
	char buffer[LCD_SIZE];
	char c;
	char c2 = -1;
	char line[100];
	char tooLong;
	
	switchToMenuMode();
	
	clearAndSetDrawBuffer(buffer);
	setFont6x8();
	
	y = 0;
	i = 0;
	j = 0;
	k = 0;
	do {
		c = s[i++];
		line[j++] = c;
		
		if (c == ' ' || c == '-' || c == '\0') {
			tooLong = j >= (c == ' ' ? 27 : 26);
			if (!tooLong && c != '\0')
				k = j;
			
			if (tooLong) {
				c2 = line[k];
				line[k] = '\0';
				drawCenter(y, line);
				y += 11;
				line[0] = c2;
				memmove(line + 1, line + k + 1, j - k - 1);
				j -= k;
			}
			
			if (c == '\0') {
				drawCenter(y, line);
				y += 11;
			}
		}
	}
	while (c != '\0');
	
	yo = (100 - (y - 3 - 1)) / 2;
	//30 = 240 / 8
	memmove(buffer + yo * 30, buffer, (y - 3) * 30);
	memset(buffer, 0, yo * 30);
	
	PortRestore();
	draw89Size(buffer);
}

//Draws the specified message, sleeps for a short time, and flushes the key
//queue.  Permits newline characters.  The text is initially drawn with the 8x10
//font.  When a '|' or a '`' is encountered, it switches to switch to a 6x8 or
//Xx6 font respectively, and adds some space.
void drawMessageAndSleep(const char *s) {
	drawMessage(s);
	sleep(500);
	flushKeyQueue();
}

//Draws a message indicating that the opponent has quit the game, and sleeps for
//a short time
void drawOpponentQuit() {
	drawMessageAndSleep("Your opponent quit`Press ESC to exit");
}

//The number of coordinates used in the "you won" graphic
#define WINNER_COORDINATES 56

//Draws the winner of a match, and sleeps for a short time.  Automatically
//switches to menu mode.
void drawWinner(short yourScore, short hisScore) {
	short i;
	//The points in the "you won" graphic
	static const unsigned char PTS[WINNER_COORDINATES] =
		{18, 17, 42, 30, 48, 7, 63, 32, 72, 9, 84, 28, 96, 5, 104, 28,
		 117, 8, 122, 28, 145, 22, 132, 43, 150, 55, 132, 64, 141, 85, 119, 77,
		 110, 92, 101, 73, 91, 92, 84, 74, 75, 90, 64, 71, 50, 90, 42, 71,
		 20, 82, 28, 63, 10, 51, 31, 43};
	char buffer[LCD_SIZE];
	char s[50];
	
	switchToMenuMode();
	
	clearAndSetDrawBuffer(buffer);
	setFont8x10();
	if (yourScore > hisScore) {
		//Draw the "you won" graphic, a jagged outline-ish figure
		//I'm shooting for a rather irregular appearance
		SetCurAttr(A_NORMAL);
		MoveTo(31, 43);
		for(i = 0; i < WINNER_COORDINATES; i += 2)
			LineTo(PTS[i], PTS[i + 1]);
		
		drawCenter(40, "You win!");
	}
	else
		drawCenter(40, "You lose.");
	
	sprintf(s, "%d - %d", yourScore, hisScore);
	setFont6x8();
	drawCenter(55, s);
	PortRestore();
	draw89Size(buffer);
	sleep(500);
	flushKeyQueue();
}















#endif
