/*
Created on August 15, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file provides a random number generator.  It is used in place of the one
used by TIGCC when predictability and careful control over each time a number is
generated by it is necessary.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef RNG_H_INCLUDED
#define RNG_H_INCLUDED

//Represents the state of a random number generator.  Its field should not be
//accessed outside this file.
typedef struct {
	unsigned long seed;
} RNG;

//Frees the contents of the specified RNG (but not the pointer to the RNG
//itself)
#define freeRNG(x)

//Returns a new RNG initialized with the specified seed
RNG *newRNG(unsigned long seed) {
	RNG *rng = (RNG *)malloc_throw(sizeof(RNG));
	rng->seed = seed;
	return rng;
}

//Returns a random unsigned long
static unsigned long rngRandLong(RNG *rng) {
	unsigned long num = (rng->seed = 69069ul * rng->seed + 1) & 0xFFFF0000;
	num |= ((rng->seed = 69069ul * rng->seed + 1) & 0xFFFF0000) / 65536l;
	return num;
}

//Returns a random number from 0 to max - 1
short rngRandom(RNG *rng, short max) {
	//If max does not evenly divide 2^32, it is slightly more probable that the
	//result will be low.  I consider this slight unevenness in distribution to
	//be trivial.
	return (short)(rngRandLong(rng) % (unsigned long)max);
}



















#endif
