/*
Created on August 9, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains a few timing methods that allow global access to a
millisecond precision (but not necessarily millisecond accuracy) timer.  Changes
the start value for AUTO_INT_5 to 254.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef TIME_H_INCLUDED
#define TIME_H_INCLUDED

//The maximum value of timerTicks
#define MAXIMUM_TIMER_TICKS 4000000000ul
//The maximum value for time().  Close to the maximum long value, but not so
//close as to be in danger of arithmetic resulting in negative numbers due to
//exceeding the maximum long value.
#define MAX_TIME 2000000000l

//The number of times AUTO_INT_5 has been executed.  It is decrimented in key.h,
//which is in charge of AUTO_INT_5.
volatile unsigned long timerTicks;

//Resets the timer
void resetTimer() {
	timerTicks = 0;
}

//The minimum value for 0x600017 (excluding 0) at the beginning of the program
unsigned char clockMin;
//The number of milliseconds per clock tick, after setting the starting value
//for 0x600017 to 254
float clockMs;

//Initializes the timer.  Must be called before calling any other timer
//functions.
void initTimer() {
	//Increase the rate at which AUTO_INT_5 is fired
	
	//while(peekIO(0x600017));
	//while(!peekIO(0x600017));
	//clockMin = peekIO(0x600017);
	//pokeIO(0x600017, 254);
	
	clockMin = PRG_getStart();
	PRG_setStart(254);
	clockMs = 3.0 / (float)(257 - (short)clockMin) * 50.0;
	resetTimer();
}

//Should be called at the end of the program if initTimer was ever called
void cleanupTimer() {
	PRG_setStart(clockMin);
}

//Returns the amount of time elapsed since the timer was initialized or last
//reset, in milliseconds, to a maximum of MAX_TIME.  MAX_TIME is close to the
//maximum long value, but not so close as to be in danger of arithmetic
//resulting in negative numbers due to exceeding the maximum long value.
long time() {
	//Note that this relies on the fact that floats have sufficient precision so
	//that rounding errors do not have a substantial impact
	return (long)min(clockMs * (float)timerTicks, MAX_TIME);
}

//Whether the timer has reached its maximum value, so that time() will
//subsequently return MAX_TIME each time
char timerExpired() {
	return time() >= MAX_TIME;
}

//Does nothing for the specified number of milliseconds
void sleep(long t) {
	long end;
	idle();
	end = min(time() + t, MAX_TIME);
	while (time() < end);
}










#endif
